package gregapi.block.prefixblock;

import static gregapi.data.CS.C;
import static gregapi.data.CS.TOOL_crowbar;
import static gregapi.data.CS.U;
import static gregapi.data.CS.W;
import gregapi.GT_API;
import gregapi.block.*;
import gregapi.code.ArrayListNoNulls;
import gregapi.data.MT;
import gregapi.data.OP;
import gregapi.data.TD;
import gregapi.item.CreativeTab;
import gregapi.lang.LanguageHandler;
import gregapi.log.GT_Log;
import gregapi.network.INetworkHandler;
import gregapi.oredict.OreDictManager;
import gregapi.oredict.OreDictMaterial;
import gregapi.oredict.OreDictMaterialStack;
import gregapi.oredict.OreDictPrefix;
import gregapi.render.*;
import gregapi.tileentity.ITileEntity;
import gregapi.util.OM;
import gregapi.util.UT;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Random;

import net.minecraft.block.Block;
import net.minecraft.block.BlockFalling;
import net.minecraft.block.ITileEntityProvider;
import net.minecraft.block.material.Material;
import net.minecraft.client.renderer.texture.IIconRegister;
import net.minecraft.creativetab.CreativeTabs;
import net.minecraft.enchantment.EnchantmentHelper;
import net.minecraft.entity.Entity;
import net.minecraft.entity.boss.EntityDragon;
import net.minecraft.entity.boss.EntityWither;
import net.minecraft.entity.item.EntityItem;
import net.minecraft.entity.player.EntityPlayer;
import net.minecraft.inventory.IInventory;
import net.minecraft.item.Item;
import net.minecraft.item.ItemStack;
import net.minecraft.nbt.NBTTagCompound;
import net.minecraft.stats.StatList;
import net.minecraft.tileentity.TileEntity;
import net.minecraft.util.IIcon;
import net.minecraft.util.StatCollector;
import net.minecraft.world.Explosion;
import net.minecraft.world.IBlockAccess;
import net.minecraft.world.World;
import net.minecraftforge.common.util.ForgeDirection;
import net.minecraftforge.event.ForgeEventFactory;
import cpw.mods.fml.common.registry.GameRegistry;

/**
 * @author Gregorius Techneticies
 */
public class PrefixBlock extends Block implements Runnable, IBlockExtendedMetaData, IBlockSyncData, IRenderedBlock, ITileEntityProvider, IBlockToolable, IBlockPlacable, IBlockRetrievable, IBlockMaterial {
	/** There are quite some odd timings in the Block breaking Code which require the TileEntity to be buffered. */
	public static ThreadLocal<PrefixBlockTileEntity> sTemporaryTileEntity = new ThreadLocal<PrefixBlockTileEntity>();
	
	public Drops mDrops;
	
	public final int mHarvestLevelOffset, mHarvestLevelMinimum, mHarvestLevelMaximum;
	public final ITexture mTexture;
	public final String mNameInternal, mTool, mModIDOwner;
	public final OreDictPrefix mPrefix;
	public final OreDictMaterialStack mHullMaterial;
	public final float mBaseHardness, mBaseResistance;
	public final boolean mGravity, mBeaconBase, mEnderDragonProof, mWitherProof, mOpaque, mNormalCube, mPlacementChecksTemperature, mPlacementChecksAntimatter, mCanBurn, mCanExplode, mRenderOverlayInWorld, mCanGlow, mCanLight;
	
	@Deprecated
	public PrefixBlock(String aModIDOwner, String aModIDTextures, String aNameInternal, OreDictPrefix aPrefix, OreDictMaterialStack aHullMaterial, Class<? extends PrefixBlockItem> aItemClass, Drops aDrops, ITexture aTexture, Material aVanillaMaterial, SoundType aSoundType, String aTool, float aBaseHardness, float aBaseResistance, int aHarvestLevelOffset, int aHarvestLevelMinimum, int aHarvestLevelMaximum, boolean aGravity, boolean aBeaconBase, boolean aEnderDragonProof, boolean aWitherProof, boolean aOpaque, boolean aNormalCube, boolean aPlacementChecksTemperature, boolean aPlacementChecksAntimatter, boolean aCanBurn, boolean aCanExplode, boolean aRenderOverlayInWorld) {
		this(aModIDOwner, aModIDTextures, aNameInternal, aPrefix, aHullMaterial, aItemClass, aDrops, aTexture, aVanillaMaterial, aSoundType, aTool, aBaseHardness, aBaseResistance, aHarvestLevelOffset, aHarvestLevelMinimum, aHarvestLevelMaximum, aGravity, aBeaconBase, aEnderDragonProof, aWitherProof, aOpaque, aNormalCube, aPlacementChecksTemperature, aPlacementChecksAntimatter, aCanBurn, aCanExplode, aRenderOverlayInWorld, true, false);
	}
	
	/**
	 * Just create one instance of this Block and everything else is getting registered automatically.
	 * 
	 * @param aModIDOwner the ID of the owning Mod. DO NOT INSERT ANY GREGTECH MODID!!!
	 * @param aModIDTextures the ID of the Texture providing Mod (for the "ModID:TextureName" thing)
	 * @param aNameInternal the internal Name of this Item. DO NOT START YOUR UNLOCALISED NAME WITH "gt."!!!
	 * @param aPrefix the OreDictPrefix corresponding to this Item.
	 * @param aHullMaterial the Material the Hull consists of. Can be null.
	 * @param aItemClass the Class of the ItemBlock to be used. If you pass null it will default to the regular MetaBlockItem Class.
	 * @param aTexture the Texture underlay for this Block. Used for Ores and Crates. Can be null to use normal Rendering.
	 * @param aVanillaMaterial the Material used to determine the Block.
	 * @param aSoundType the Sound Type of the Block.
	 * @param aTool the Tool used to harvest this Block.
	 * @param aBaseHardness if smaller than zero, then this Block is indestructible.
	 * @param aBaseResistance
	 * @param aHarvestLevelOffset
	 * @param aHarvestLevelMinimum
	 * @param aHarvestLevelMaximum
	 * @param aGravity if this Block falls like Gravel.
	 * @param aBeaconBase if this Block can be used as Beacon Base.
	 * @param aEnderDragonProof if this Block cannot be destroyed by an Ender Dragon (used for the End Ores).
	 * @param aWitherProof if this Block cannot be destroyed by a Wither.
	 * @param aOpaque if this Block is Opaque.
	 * @param aNormalCube if this Block is a normal Cube (for Redstone Stuff).
	 * @param aPlacementChecksTemperature if this Block checks for Temperature to be proper before placing.
	 * @param aPlacementChecksAntimatter if this Block checks for being Antimatter before placing.
	 * @param aCanBurn if this Block can burn if the Material it is made of can burn.
	 * @param aCanExplode if this Block can explode if the Material it is made of can explode.
	 * @param aRenderOverlayInWorld if the Icon Overlay is to be rendered InWorld. Used for Crates and Ores.
	 */
	public PrefixBlock(String aModIDOwner, String aModIDTextures, String aNameInternal, OreDictPrefix aPrefix, OreDictMaterialStack aHullMaterial, Class<? extends PrefixBlockItem> aItemClass, Drops aDrops, ITexture aTexture, Material aVanillaMaterial, SoundType aSoundType, String aTool, float aBaseHardness, float aBaseResistance, int aHarvestLevelOffset, int aHarvestLevelMinimum, int aHarvestLevelMaximum, boolean aGravity, boolean aBeaconBase, boolean aEnderDragonProof, boolean aWitherProof, boolean aOpaque, boolean aNormalCube, boolean aPlacementChecksTemperature, boolean aPlacementChecksAntimatter, boolean aCanBurn, boolean aCanExplode, boolean aRenderOverlayInWorld, boolean aCanGlow, boolean aCanLight) {
		super(aVanillaMaterial);
		mPrefix = aPrefix;
		mNameInternal = aNameInternal;
		GameRegistry.registerBlock(this, aItemClass==null?PrefixBlockItem.class:aItemClass, mNameInternal);
		
		setStepSound(aSoundType);
		mOpaque = aOpaque;
		mGravity = aGravity;
		mCanBurn = aCanBurn;
		mCanGlow = aCanGlow;
		mCanLight = aCanLight;
		mCanExplode = aCanExplode;
		mNormalCube = aNormalCube;
		mBeaconBase = aBeaconBase;
		mWitherProof = aWitherProof;
		mEnderDragonProof = aEnderDragonProof;
		mRenderOverlayInWorld = aRenderOverlayInWorld;
		mPlacementChecksAntimatter = aPlacementChecksAntimatter;
		mPlacementChecksTemperature = aPlacementChecksTemperature;
		
		mTool = aTool.toLowerCase();
		mDrops = aDrops==null?new Drops(this, this):aDrops;
		mTexture = aTexture;
		mModIDOwner = aModIDOwner;
		mHullMaterial = aHullMaterial;
		mBaseHardness = aBaseHardness;
		mBaseResistance = aBaseResistance;
		mHarvestLevelOffset = aHarvestLevelOffset;
		mHarvestLevelMinimum = Math.max(0, aHarvestLevelMinimum);
		mHarvestLevelMaximum = Math.max(aHarvestLevelMinimum, aHarvestLevelMaximum);
		mPrefix.addTextureSet(aModIDTextures, false);
		LanguageHandler.add(mNameInternal+"."+0+".name", getLocalName(mPrefix, MT.Empty));
		LanguageHandler.add(mNameInternal+"."+W+".name", "Any Sub-Block of this one"); // Local Name for the WildcardItem Variant.
		
        opaque = this.isOpaqueCube();
        lightOpacity = this.isOpaqueCube() ? 255 : 0;
        
        if (mPrefix.mCreativeTab == null) mPrefix.mCreativeTab = new CreativeTab(mPrefix.mNameInternal, mPrefix.mNameCategory, Item.getItemFromBlock(this), W);
		setCreativeTab(mPrefix.mCreativeTab);
		
        try {ic2.api.tile.ExplosionWhitelist.addWhitelistedBlock(this);} catch(Throwable e) {/**/}
        
		GT_API.sAPIInitBefore.add(this);
	}
	
	/** This ensures, that all Materials are registered at the time this Item registers to the OreDictionary. */
	@Override
	public void run() {
		boolean tUnificationAllowed = (mPrefix.contains(TD.Prefix.UNIFICATABLE) && !mPrefix.contains(TD.Prefix.UNIFICATABLE_RECIPES));
		for (short i = 0; i < OreDictMaterial.MATERIAL_ARRAY.length; i++) if (OreDictMaterial.MATERIAL_ARRAY[i] != null && mPrefix.isGeneratingItem(OreDictMaterial.MATERIAL_ARRAY[i])) {
			ItemStack tStack = new ItemStack(this, 1, i);
			UT.Stacks.update(tStack);
			LanguageHandler.add(mNameInternal+"."+i+".name", getLocalName(mPrefix, OreDictMaterial.MATERIAL_ARRAY[i]));
			if (tUnificationAllowed) OreDictManager.INSTANCE.setTarget(mPrefix, OreDictMaterial.MATERIAL_ARRAY[i], tStack); else OreDictManager.INSTANCE.registerOre(mPrefix, OreDictMaterial.MATERIAL_ARRAY[i], tStack);
		}
	}
	
	@Override
    public void registerBlockIcons(IIconRegister aIconRegister) {
		if (mPrefix.mIconIndexBlock >= 0) {
			MT.NULL.mTextureSetsBlock.get(mPrefix.mIconIndexBlock).registerIcons(aIconRegister);
			HashSet<IIconContainer> tSet = new HashSet<IIconContainer>();
			for (int i = 0; i < OreDictMaterial.MATERIAL_ARRAY.length; i++) if (OreDictMaterial.MATERIAL_ARRAY[i] != null && OreDictMaterial.MATERIAL_ARRAY[i].mTextureSetsBlock != null) {
				IIconContainer tIcon = OreDictMaterial.MATERIAL_ARRAY[i].mTextureSetsBlock.get(mPrefix.mIconIndexBlock);
				if (!tSet.contains(tIcon)) {tIcon.registerIcons(aIconRegister); tSet.add(tIcon);}
			}
		}
    }
	
    @Override
	public IIcon getIcon(IBlockAccess aWorld, int aX, int aY, int aZ, int aSide) {
        return getIcon(aSide, getMetaDataValue(aWorld, aX, aY, aZ));
    }
    
    @Override
    public IIcon getIcon(int aSide, int aMetaData) {
		if (mPrefix.mIconIndexBlock >= 0) {
			OreDictMaterial aMaterial = getMetaMaterial(aMetaData);
			if (aMaterial != null && aMaterial.mTextureSetsBlock != null)
			return aMaterial	.mTextureSetsBlock.get(mPrefix.mIconIndexBlock).getIcon(0);
			return MT.NULL		.mTextureSetsBlock.get(mPrefix.mIconIndexBlock).getIcon(0);
		}
        return null;
    }
    
    @Override
    public int getRenderColor(int aMetaData) {
		OreDictMaterial aMaterial = getMetaMaterial(aMetaData);
	    return aMaterial == null ? super.getRenderColor(aMetaData) : UT.Code.getRGBInt(aMaterial.mRGBa[mPrefix.mState]);
    }
    
	public ITexture getTexture(short aMetaData, boolean aRendersInWorld) {
		if (!mRenderOverlayInWorld && aRendersInWorld) return mTexture;
		if (mPrefix.mIconIndexBlock >= 0) {
			OreDictMaterial aMaterial = getMetaMaterial(aMetaData);
			if (mTexture == null) {
				if (aMaterial != null && aMaterial.mTextureSetsBlock != null)
				return new BlockTextureDefault(aMaterial	.mTextureSetsBlock.get(mPrefix.mIconIndexBlock), aMaterial.mRGBa[mPrefix.mState], !mOpaque, mCanGlow && aMaterial.contains(TD.Properties.GLOWING));
				return new BlockTextureDefault(MT.NULL		.mTextureSetsBlock.get(mPrefix.mIconIndexBlock));
			}
			if (aMaterial != null && aMaterial.mTextureSetsBlock != null)
			return new BlockTextureMulti(mTexture, new BlockTextureDefault(aMaterial	.mTextureSetsBlock.get(mPrefix.mIconIndexBlock), aMaterial.mRGBa[mPrefix.mState], !mOpaque, mCanGlow && aMaterial.contains(TD.Properties.GLOWING)));
			return new BlockTextureMulti(mTexture, new BlockTextureDefault(MT.NULL		.mTextureSetsBlock.get(mPrefix.mIconIndexBlock)));
		}
        return null;
	}
	
	@Override
	public ITexture getTexture(int aRenderPass, byte aSide, short aMetaData) {
        return getTexture(aMetaData, false);
	}
	
	@Override
	public ITexture getTexture(int aRenderPass, byte aSide, IBlockAccess aWorld, int aX, int aY, int aZ) {
		return getTexture(getMetaDataValue(aWorld, aX, aY, aZ), true);
	}
	
	@Override
	public boolean setBlockBounds(int aRenderPass, byte aSide, short aMetaData) {
		return false;
	}
	
	@Override
	public boolean setBlockBounds(int aRenderPass, byte aSide, IBlockAccess aWorld, int aX, int aY, int aZ) {
		return false;
	}
	
	@Override
	public int getRenderPasses(byte aSide, short aMetaData) {
		return 1;
	}
	
	@Override
	public int getRenderPasses(byte aSide, IBlockAccess aWorld, int aX, int aY, int aZ) {
		return 1;
	}
	
	@Override
	public IRenderedBlockObject passRenderingToObject(IBlockAccess aWorld, int aX, int aY, int aZ) {
		TileEntity tTileEntity = aWorld.getTileEntity(aX, aY, aZ);
		return tTileEntity instanceof IRenderedBlockObject ? (IRenderedBlockObject)tTileEntity : null;
	}
	
	@Override
	public IRenderedBlockObject passRenderingToObject(short aMetaData) {
		return null;
	}
	
	private static boolean LOCK = false;
	
	@Override
	public void onNeighborChange(IBlockAccess aWorld, int aX, int aY, int aZ, int aTileX, int aTileY, int aTileZ) {
		if (!LOCK) {
			LOCK = true;
		    TileEntity aTileEntity = aWorld.getTileEntity(aX, aY, aZ);
		    if (aTileEntity instanceof ITileEntity) ((ITileEntity)aTileEntity).onAdjacentBlockChange(aTileX, aTileY, aTileZ);
			LOCK = false;
		}
	}
	
    @Override
	public void onNeighborBlockChange(World aWorld, int aX, int aY, int aZ, Block aBlock) {
		if (!LOCK) {
			LOCK = true;
			TileEntity aTileEntity = aWorld.getTileEntity(aX, aY, aZ);
		    if (aTileEntity instanceof ITileEntity) ((ITileEntity)aTileEntity).onAdjacentBlockChange(aX, aY, aZ);
			LOCK = false;
		}
        aWorld.scheduleBlockUpdate(aX, aY, aZ, this, tickRate(aWorld));
    }
    
    @Override
    public void onBlockExploded(World aWorld, int aX, int aY, int aZ, Explosion aExplosion) {
    	if (aWorld.isRemote) return;
		OreDictMaterial aMaterial = getMetaMaterial(aWorld, aX, aY, aZ);
		aWorld.setBlockToAir(aX, aY, aZ);
		if (aMaterial != null && ((mCanExplode && aMaterial.contains(TD.Properties.EXPLOSIVE)) || (mCanBurn && aMaterial.contains(TD.Properties.FLAMMABLE) && mPrefix.mFamiliarPrefixes.contains(OP.dust)))) try {aWorld.newExplosion(null, aX+0.5, aY+0.5, aZ+0.5, (mPrefix.mAmount*0.7F)/U, true, true);} catch(StackOverflowError e) {GT_Log.err.println("WARNING: StackOverflow during Explosion has been prevented at: " + aX +" ; "+ aY +" ; "+ aZ);}
    }
    
    @Override
	public boolean onBlockEventReceived(World aWorld, int aX, int aY, int aZ, int aID, int aData) {
        TileEntity aTileEntity = aWorld.getTileEntity(aX, aY, aZ);
        return aTileEntity == null || aTileEntity.receiveClientEvent(aID, aData);
    }
    
	@Override
    public int getDamageValue(World aWorld, int aX, int aY, int aZ) {
	    return getMetaDataValue(aWorld, aX, aY, aZ);
    }
	
	@Override
	public void breakBlock(World aWorld, int aX, int aY, int aZ, Block aBlock, int par6) {
	    TileEntity tTileEntity = aWorld.getTileEntity(aX, aY, aZ);
	    if (tTileEntity instanceof PrefixBlockTileEntity) sTemporaryTileEntity.set((PrefixBlockTileEntity)tTileEntity);
	    aWorld.removeTileEntity(aX, aY, aZ);
	}
	
	@Override
	public boolean placeBlock(World aWorld, int aX, int aY, int aZ, byte aSide, short aMetaData, NBTTagCompound aNBT, boolean aCauseBlockUpdates, boolean aForcePlacement) {
		OreDictMaterial aMaterial = getMetaMaterial(aMetaData);
		if (aMaterial != null && (aForcePlacement || ((!mPlacementChecksAntimatter || !aMaterial.contains(TD.Atomic.ANTIMATTER)) && (!mPlacementChecksTemperature || aMaterial.mMeltingPoint > UT.Worlds.getTemperature(aWorld, aX, aY, aZ)))) && aWorld.setBlock(aX, aY, aZ, this, (int)UT.Code.bind(0, 15, aMaterial.mToolQuality), aCauseBlockUpdates?3:0)) {
		    // This darn TileEntity update is ruining World generation Code (infinite Loops when placing TileEntities on Chunk Borders). I'm glad I finally found a way to disable it.
			UT.Worlds.setTileEntity(aWorld, aX, aY, aZ, createTileEntity(aWorld, aX, aY, aZ, aSide, aMetaData, aNBT), aCauseBlockUpdates);
	        return true;
        }
		return false;
	}
	
	@Override
	public ItemStack getItemStackFromBlock(IBlockAccess aWorld, int aX, int aY, int aZ, byte aSide) {
	    return new ItemStack(this, 1, getMetaDataValue(aWorld, aX, aY, aZ));
	}
	
	@Override
	public int getFlammability(IBlockAccess aWorld, int aX, int aY, int aZ, ForgeDirection aSide) {
		OreDictMaterialStack aMaterial = getMaterialAtSide(aWorld, aX, aY, aZ, (byte)aSide.ordinal());
		return aMaterial == null || !mCanBurn ? 0 : (aMaterial.mMaterial.contains(TD.Properties.FLAMMABLE)?100:0) + (aMaterial.mMaterial.contains(TD.Properties.BURNING)?200:0);
	}
	
	@Override
	public int getFireSpreadSpeed(IBlockAccess aWorld, int aX, int aY, int aZ, ForgeDirection aSide) {
		OreDictMaterialStack aMaterial = getMaterialAtSide(aWorld, aX, aY, aZ, (byte)aSide.ordinal());
		return aMaterial == null || !mCanBurn ? 0 : (aMaterial.mMaterial.contains(TD.Properties.FLAMMABLE)?100:0) + (aMaterial.mMaterial.contains(TD.Properties.BURNING)?200:0);
	}
	
	@Override
	public boolean isFireSource(World aWorld, int aX, int aY, int aZ, ForgeDirection aSide) {
		OreDictMaterialStack aMaterial = getMaterialAtSide(aWorld, aX, aY, aZ, (byte)aSide.ordinal());
		return aMaterial != null && mCanBurn && aMaterial.mMaterial.contains(TD.Properties.FLAMMABLE) && aMaterial.mMaterial.contains(TD.Properties.UNBURNABLE);
	}
	
	@Override
	public boolean canEntityDestroy(IBlockAccess aWorld, int aX, int aY, int aZ, Entity aEntity) {
		if (aEntity instanceof EntityDragon) {
			if (mEnderDragonProof) return false;
			OreDictMaterialStack aMaterial = getMaterialAtSide(aWorld, aX, aY, aZ, (byte)6);
			return aMaterial == null || !aMaterial.mMaterial.contains(TD.Properties.ENDER_DRAGON_PROOF);
		}
		if (aEntity instanceof EntityWither) {
			if (mWitherProof) return false;
			OreDictMaterialStack aMaterial = getMaterialAtSide(aWorld, aX, aY, aZ, (byte)6);
			return aMaterial == null || !aMaterial.mMaterial.contains(TD.Properties.WITHER_PROOF);
		}
		return true;
	}
	
	@Override
	public long onToolClick(String aTool, long aRemainingDurability, Entity aPlayer, IInventory aPlayerInventory, boolean aSneaking, ItemStack aStack, World aWorld, byte aSide, int aX, int aY, int aZ, float aHitX, float aHitY, float aHitZ) {
		// TODO: I directly inserted the Crate Drop Code here, instead of making a special registry.
		if (!aWorld.isRemote && aTool.equals(TOOL_crowbar)) {
			OreDictMaterial aMaterial = getMetaMaterial(aWorld, aX, aY, aZ);
			if (mPrefix == OP.crateGtGem) {
				UT.Sounds.send(aWorld, "random.break", 1.0F, -1.0F, aX, aY, aZ);
				aWorld.setBlockToAir(aX, aY, aZ);
				aWorld.spawnEntityInWorld(new EntityItem(aWorld, aX+0.5, aY+0.5, aZ+0.5, OM.get(OP.gem, aMaterial, 16)));
				return 10000;
			}
			if (mPrefix == OP.crateGtDust) {
				UT.Sounds.send(aWorld, "random.break", 1.0F, -1.0F, aX, aY, aZ);
				aWorld.setBlockToAir(aX, aY, aZ);
				aWorld.spawnEntityInWorld(new EntityItem(aWorld, aX+0.5, aY+0.5, aZ+0.5, OM.get(OP.dust, aMaterial, 16)));
				return 10000;
			}
			if (mPrefix == OP.crateGtIngot) {
				UT.Sounds.send(aWorld, "random.break", 1.0F, -1.0F, aX, aY, aZ);
				aWorld.setBlockToAir(aX, aY, aZ);
				aWorld.spawnEntityInWorld(new EntityItem(aWorld, aX+0.5, aY+0.5, aZ+0.5, OM.get(OP.ingot, aMaterial, 16)));
				return 10000;
			}
			if (mPrefix == OP.crateGtPlate) {
				UT.Sounds.send(aWorld, "random.break", 1.0F, -1.0F, aX, aY, aZ);
				aWorld.setBlockToAir(aX, aY, aZ);
				aWorld.spawnEntityInWorld(new EntityItem(aWorld, aX+0.5, aY+0.5, aZ+0.5, OM.get(OP.plate, aMaterial, 16)));
				return 10000;
			}
			if (mPrefix == OP.crateGtPlateGem) {
				UT.Sounds.send(aWorld, "random.break", 1.0F, -1.0F, aX, aY, aZ);
				aWorld.setBlockToAir(aX, aY, aZ);
				aWorld.spawnEntityInWorld(new EntityItem(aWorld, aX+0.5, aY+0.5, aZ+0.5, OM.get(OP.plateGem, aMaterial, 16)));
				return 10000;
			}
		}
		// Proceed with the regular onToolClick of the ToolCompat Class, because it has important Code in it.
		return ToolCompat.onToolClick(this, aTool, aRemainingDurability, aPlayer, aPlayerInventory, aSneaking, aStack, aWorld, aSide, aX, aY, aZ, aHitX, aHitY, aHitZ);
	}
	
	@Override
	public OreDictMaterialStack getMaterialAtSide(IBlockAccess aWorld, int aX, int aY, int aZ, byte aSide) {
		if (mHullMaterial != null) return mHullMaterial;
		OreDictMaterial aMaterial = getMetaMaterial(aWorld, aX, aY, aZ);
		return aMaterial == null ? null : OM.stack(mPrefix, aMaterial);
	}
	
	@Override
	public void setExtendedMetaData(IBlockAccess aWorld, int aX, int aY, int aZ, short aMetaData) {
		TileEntity aTileEntity = aWorld.getTileEntity(aX, aY, aZ);
		if (aTileEntity == null && aWorld instanceof World) aTileEntity = UT.Worlds.setTileEntity((World)aWorld, aX, aY, aZ, createTileEntity((World)aWorld, aX, aY, aZ, (byte)6, aMetaData, null), false);
		if (aTileEntity instanceof PrefixBlockTileEntity) ((PrefixBlockTileEntity)aTileEntity).mMetaData = aMetaData;
		if (aWorld instanceof World && ((World)aWorld).isRemote) ((World)aWorld).markBlockForUpdate(aX, aY, aZ);
	}
	
	@Override
	public short getExtendedMetaData(IBlockAccess aWorld, int aX, int aY, int aZ) {
		return getMetaDataValue(aWorld, aX, aY, aZ);
	}
	
	@Override
	public boolean removeMaterialFromSide(World aWorld, int aX, int aY, int aZ, byte aSide, OreDictMaterialStack aMaterial) {
		OreDictMaterialStack tMaterial = getMaterialAtSide(aWorld, aX, aY, aZ, aSide);
		if (aMaterial.mMaterial == tMaterial.mMaterial && aMaterial.mAmount > 0 && aMaterial.mAmount <= tMaterial.mAmount) {
			ItemStack tStack = OM.dust(OM.stack(tMaterial.mAmount - aMaterial.mAmount, aMaterial.mMaterial));
			if (tStack != null) aWorld.spawnEntityInWorld(new EntityItem(aWorld, aX+0.5, aY+0.5, aZ+0.5, tStack));
			aWorld.setBlockToAir(aX, aY, aZ);
			return true;
		}
		return false;
	}
	
    @Override
    public void updateTick(World aWorld, int aX, int aY, int aZ, Random aRandom) {
    	if (aWorld.isRemote || checkGravity(aWorld, aX, aY, aZ)) return;
		OreDictMaterial aMaterial = getMetaMaterial(aWorld, aX, aY, aZ);
		if (aMaterial != null && mCanBurn && (mPrefix.mFamiliarPrefixes.contains(OP.dust) || (mCanExplode && aMaterial.contains(TD.Properties.EXPLOSIVE))) && aMaterial.contains(TD.Properties.FLAMMABLE) && UT.Worlds.getTemperature(aWorld, aX, aY, aZ) > C + 100) {
			aWorld.setBlockToAir(aX, aY, aZ);
			try {aWorld.newExplosion(null, aX+0.5, aY+0.5, aZ+0.5, (aMaterial.contains(TD.Properties.EXPLOSIVE)?mPrefix.mAmount*0.5F:mPrefix.mAmount*0.33F)/U, true, true);} catch(StackOverflowError e) {GT_Log.err.println("WARNING: StackOverflow during Explosion has been prevented at: " + aX +" ; "+ aY +" ; "+ aZ);}
		}
    }
    
    @Override
    public void dropBlockAsItemWithChance(World aWorld, int aX, int aY, int aZ, int aMeta, float aChance, int aFortune) {
        ArrayList<ItemStack> tList = mDrops.getDrops(this, aWorld, aX, aY, aZ, aFortune, false);
        aChance = ForgeEventFactory.fireBlockHarvesting(tList, aWorld, this, aX, aY, aZ, aMeta, aFortune, aChance, false, harvesters.get());
        for (ItemStack tStack : tList) if (aWorld.rand.nextFloat() <= aChance) dropBlockAsItem(aWorld, aX, aY, aZ, tStack);
    }
    
    @Override
    public void harvestBlock(World aWorld, EntityPlayer aPlayer, int aX, int aY, int aZ, int aMeta) {
        aPlayer.addStat(StatList.mineBlockStatArray[getIdFromBlock(this)], 1);
        aPlayer.addExhaustion(0.025F);
	    boolean aSilkTouch = EnchantmentHelper.getSilkTouchModifier(aPlayer);
	    int aFortune = EnchantmentHelper.getFortuneModifier(aPlayer);
	    ArrayList<ItemStack> tList = mDrops.getDrops(this, aWorld, aX, aY, aZ, aFortune, aSilkTouch);
	    float aChance = ForgeEventFactory.fireBlockHarvesting(tList, aWorld, this, aX, aY, aZ, aMeta, aFortune, 1.0F, aSilkTouch, aPlayer);
	    for (ItemStack tStack : tList) if (aWorld.rand.nextFloat() <= aChance) dropBlockAsItem(aWorld, aX, aY, aZ, tStack);
    }
    
    @Override public int getRenderBlockPass() {return 0;}
    @Override public final ArrayList<ItemStack> getDrops(World aWorld, int aX, int aY, int aZ, int aUnusableMetaData, int aFortune) {return mDrops.getDrops(this, aWorld, aX, aY, aZ, aFortune, false);}
    @Override public void onBlockAdded(World aWorld, int aX, int aY, int aZ) {aWorld.scheduleBlockUpdate(aX, aY, aZ, this, tickRate(aWorld));}
    @Override public void getSubBlocks(Item aItem, CreativeTabs aCreativeTab, List aList) {aItem.getSubItems(aItem, aCreativeTab, aList);}
    /** Where I come from, we set the TileEntities ourselves instead of letting a Handler do it. */
	@Override public final TileEntity createNewTileEntity(World aWorld, int aMeta) {return null;}
    /** Where I come from, we set the TileEntities ourselves instead of letting a Handler do it. */
	@Override public final TileEntity createTileEntity(World aWorld, int aMeta) {return null;}
	@Override public String getUnlocalizedName() {return mNameInternal;}
	@Override public String getLocalizedName() {return StatCollector.translateToLocal(mNameInternal + ".name");}
	@Override public String getHarvestTool(int aMaterialToolQuality) {return mTool;}
	@Override public float getBlockHardness(World aWorld, int aX, int aY, int aZ)																						{return mBaseHardness < 0 ? -1 : mBaseHardness * (1+getHarvestLevel(aWorld.getBlockMetadata(aX, aY, aZ)));}
	@Override public float getExplosionResistance(Entity par1Entity, World aWorld, int aX, int aY, int aZ, double explosionX, double explosionY, double explosionZ)		{return mBaseResistance * (1+getHarvestLevel(aWorld.getBlockMetadata(aX, aY, aZ)));}
	@Override public int getRenderType() {return RendererBlockTextured.INSTANCE==null?super.getRenderType():RendererBlockTextured.INSTANCE.mRenderID;}
	@Override public int getHarvestLevel(int aMaterialToolQuality) {return (int)UT.Code.bind(mHarvestLevelMinimum, mHarvestLevelMaximum, mHarvestLevelOffset + aMaterialToolQuality);}
    @Override public int tickRate(World aWorld) {return 2;}
	@Override public int colorMultiplier(IBlockAccess aWorld, int aX, int aY, int aZ) {return getRenderColor(getMetaDataValue(aWorld, aX, aY, aZ));}
	@Override public int getLightOpacity() {return mOpaque?255:0;}
	@Override public boolean isBeaconBase(IBlockAccess aWorld, int aX, int aY, int aZ, int aBeaconX, int aBeaconY, int aBeaconZ) {return mBeaconBase;}
    @Override public boolean isSideSolid(IBlockAccess aWorld, int aX, int aY, int aZ, ForgeDirection aSide) {return mOpaque;}
	@Override public boolean canBeReplacedByLeaves(IBlockAccess aWorld, int aX, int aY, int aZ) {return false;}
	@Override public boolean isNormalCube(IBlockAccess aWorld, int aX, int aY, int aZ)  {return mNormalCube;}
	@Override public boolean hasTileEntity(int aMeta) {return true;}
	@Override public boolean renderAsNormalBlock() {return true;}
	@Override public boolean isOpaqueCube() {return mOpaque;}
	@Override public boolean canSilkHarvest() {return false;}
	@Override public boolean func_149730_j() {return mOpaque;}
	
	/** @return the Local Name for this Block depending on Prefix and Material. */
	public String getLocalName(OreDictPrefix aPrefix, OreDictMaterial aMaterial) {
		return LanguageHandler.getLocalName(aPrefix, aMaterial);
	}
	
	public short getMetaDataValue(TileEntity aTileEntity) {
	    return aTileEntity instanceof PrefixBlockTileEntity?((PrefixBlockTileEntity)aTileEntity).mMetaData:0;
    }
    
	public short getMetaDataValue(IBlockAccess aWorld, int aX, int aY, int aZ) {
	    return getMetaDataValue(aWorld.getTileEntity(aX, aY, aZ));
    }
    
	public OreDictMaterial getMetaMaterial(int aMetaData) {
		return UT.Code.exists(aMetaData, OreDictMaterial.MATERIAL_ARRAY)?OreDictMaterial.MATERIAL_ARRAY[aMetaData]:null;
    }
    
	public OreDictMaterial getMetaMaterial(TileEntity aTileEntity) {
	    return getMetaMaterial(aTileEntity instanceof PrefixBlockTileEntity?((PrefixBlockTileEntity)aTileEntity).mMetaData:0);
    }
    
	public OreDictMaterial getMetaMaterial(IBlockAccess aWorld, int aX, int aY, int aZ) {
		return getMetaMaterial(aWorld.getTileEntity(aX, aY, aZ));
    }
	
	public TileEntity createTileEntity(World aWorld, int aX, int aY, int aZ, byte aSide, short aMetaData, NBTTagCompound aNBT) {
		PrefixBlockTileEntity rTileEntity = new PrefixBlockTileEntity();
		if (aNBT != null) rTileEntity.readFromNBT(aNBT);
		rTileEntity.mMetaData = aMetaData;
		return rTileEntity;
    }
    
    protected boolean checkGravity(World aWorld, int aX, int aY, int aZ) {
        if (mGravity && aY > 0 && aWorld.getTileEntity(aX, aY, aZ) != null && BlockFalling.func_149831_e(aWorld, aX, aY - 1, aZ)) {
            if (!BlockFalling.fallInstantly && aWorld.checkChunksExist(aX-32, aY-32, aZ-32, aX+32, aY+32, aZ+32)) {
                if (!aWorld.isRemote) aWorld.spawnEntityInWorld(new PrefixBlockFallingEntity(aWorld, aX+0.5, aY+0.5, aZ+0.5, this, getItemStackFromBlock(aWorld, aX, aY, aZ, (byte)1)));
            } else {
            	short tMetaData = getMetaDataValue(aWorld, aX, aY, aZ);
                aWorld.setBlockToAir(aX, aY, aZ);
                while (BlockFalling.func_149831_e(aWorld, aX, aY-1, aZ) && aY > 0) --aY;
                if (aY > 0) placeBlock(aWorld, aX, aY, aZ, (byte)1, tMetaData, null, false, true);
            }
            return true;
        }
        return false;
    }
    
    public static class Drops {
    	private final Item mDropNormal, mDropSilkTouch;
    	
    	public Drops(Item  aDropNormal) {this(aDropNormal, aDropNormal);}
    	public Drops(Block aDropNormal) {this(Item.getItemFromBlock(aDropNormal));}
    	public Drops(Block aDropNormal, Block aDropSilkTouch) {this(Item.getItemFromBlock(aDropNormal), Item.getItemFromBlock(aDropSilkTouch));}
    	public Drops(Item  aDropNormal, Block aDropSilkTouch) {this(aDropNormal, Item.getItemFromBlock(aDropSilkTouch));}
    	public Drops(Block aDropNormal, Item  aDropSilkTouch) {this(Item.getItemFromBlock(aDropNormal), aDropSilkTouch);}
    	
    	public Drops(Item aDropNormal, Item aDropSilkTouch) {
    		mDropNormal = aDropNormal;
    		mDropSilkTouch = aDropSilkTouch;
    	}
    	
    	public ArrayList<ItemStack> getDrops(PrefixBlock aBlock, World aWorld, int aX, int aY, int aZ, int aFortune, boolean aSilkTouch) {
    	    TileEntity aTileEntity = aWorld.getTileEntity(aX, aY, aZ);
    		if (aTileEntity == null) {
    			aTileEntity = sTemporaryTileEntity.get();
        		if (aTileEntity == null || aTileEntity.xCoord != aX || aTileEntity.yCoord != aY || aTileEntity.zCoord != aZ) aTileEntity = null;
    		}
            return getDrops(aBlock, aWorld, aX, aY, aZ, aBlock.getMetaDataValue(aTileEntity), aFortune, aSilkTouch);
        }
    	
    	public ArrayList<ItemStack> getDrops(PrefixBlock aBlock, World aWorld, int aX, int aY, int aZ, short aMetaData, int aFortune, boolean aSilkTouch) {
    	    ArrayListNoNulls<ItemStack> rList = new ArrayListNoNulls<ItemStack>();
    	    rList.add(new ItemStack(aSilkTouch?mDropSilkTouch:mDropNormal, 1, aMetaData));
            return rList;
        }
    }
    
    public static class Drops_None extends Drops {
    	public Drops_None() {
    		super((Item)null);
		}
    	
    	@Override
		public ArrayList<ItemStack> getDrops(PrefixBlock aBlock, World aWorld, int aX, int aY, int aZ, short aMetaData, int aFortune, boolean aSilkTouch) {
    		return new ArrayListNoNulls<ItemStack>();
        }
    }
    
    public static class Drops_Small_Ore extends Drops {
    	private final OreDictMaterial mSecondaryDrop;
    	
    	public Drops_Small_Ore(OreDictMaterial aSecondaryDrop) {
    		super((Item)null);
    		mSecondaryDrop = aSecondaryDrop;
		}
    	
    	@Override
		public ArrayList<ItemStack> getDrops(PrefixBlock aBlock, World aWorld, int aX, int aY, int aZ, short aMetaData, int aFortune, boolean aSilkTouch) {
    		ArrayList<ItemStack> rList = new ArrayListNoNulls<ItemStack>();
    	    OreDictMaterial aMaterial = aBlock.getMetaMaterial(aMetaData);
            
    		if (aMaterial != null) {
    			Random tRandom = new Random(aX ^ aY ^ aZ);
    			ArrayList<ItemStack> tSelector = new ArrayListNoNulls<ItemStack>();
    			ItemStack
    			
    			tStack = OM.get(OP.gemExquisite, aMaterial, OM.get(OP.gem, aMaterial, 1), 1);
    			if (tStack != null) for (byte i = 0; i <  1; i++) tSelector.add(tStack);
    			tStack = OM.get(OP.gemFlawless, aMaterial, OM.get(OP.gem, aMaterial, 1), 1);
    			if (tStack != null) for (byte i = 0; i <  2; i++) tSelector.add(tStack);
    			tStack = OM.get(OP.gem, aMaterial, 1);
    			if (tStack != null) for (byte i = 0; i < 12; i++) tSelector.add(tStack);
    			
    			tStack = OM.get(OP.gemFlawed, aMaterial, OM.get(OP.crushed, aMaterial, 1), 1);
    			if (tStack != null) for (byte i = 0; i <  5; i++) tSelector.add(tStack);
    			tStack = OM.get(OP.crushed, aMaterial, 1);
    			if (tStack != null) for (byte i = 0; i < 10; i++) tSelector.add(tStack);
    			
    			tStack = OM.get(OP.gemChipped, aMaterial, OM.get(OP.dustImpure, aMaterial, 1), 1);
    			if (tStack != null) for (byte i = 0; i <  5; i++) tSelector.add(tStack);
    			tStack = OM.get(OP.dustImpure, aMaterial, OM.get(OP.dust, aMaterial, 1), 1);
    			if (tStack != null) for (byte i = 0; i < 10; i++) tSelector.add(tStack);
    			
    			if (tSelector.size() > 0) {
    				for (int i = 0, j = Math.max(1, aMaterial.mOreMultiplier+(aFortune>0?(tRandom.nextInt((1+aFortune)*aMaterial.mOreMultiplier)):0)/2); i < j; i++) {
    					rList.add(UT.Stacks.copyAmount(1, tSelector.get(tRandom.nextInt(tSelector.size()))));
    				}
    			}
    			
    			if (mSecondaryDrop != null && tRandom.nextInt(3+aFortune)>1) rList.add(OM.get(tRandom.nextInt(3)>0?OP.dustImpure:OP.dust, mSecondaryDrop.mTargetCrushing.mMaterial, 1));
    		}
    		return rList;
        }
    }
    
	@Override public void receiveDataByte				(IBlockAccess aWorld, int aX, int aY, int aZ, byte   aData, INetworkHandler aNetworkHandler) {/**/}
	@Override public void receiveDataShort				(IBlockAccess aWorld, int aX, int aY, int aZ, short  aData, INetworkHandler aNetworkHandler) {setExtendedMetaData(aWorld, aX, aY, aZ, aData);}
	@Override public void receiveDataInteger			(IBlockAccess aWorld, int aX, int aY, int aZ, int    aData, INetworkHandler aNetworkHandler) {/**/}
	@Override public void receiveDataLong				(IBlockAccess aWorld, int aX, int aY, int aZ, long   aData, INetworkHandler aNetworkHandler) {/**/}
	@Override public void receiveDataByteArray			(IBlockAccess aWorld, int aX, int aY, int aZ, byte[] aData, INetworkHandler aNetworkHandler) {/**/}
}